from importlib import import_module

from django.core.exceptions import ImproperlyConfigured


def _get_module_method(path_name):
    if not path_name:
        return
    module_name, method_name = path_name.rsplit(".", 1)
    try:
        mod = import_module(module_name)
    except ImportError as e:
        raise ImproperlyConfigured(
            ('Error importing module %s: "%s"' % (module_name, e))
        )
    return getattr(mod, method_name)



import functools
import hashlib
from urllib.parse import quote

from django.core.cache import cache
from django.utils.encoding import force_bytes


def make_fragment_key(fragment_key: str, *vary_on):
    key = ":".join(quote(str(var)) for var in vary_on)
    args = hashlib.md5(force_bytes(key))
    return fragment_key % (args.hexdigest())


def cache_func(ttl):
    def decorator_cache(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cache_key = make_fragment_key(
                f"{func.__module__}.{func.__name__}.%s", args, kwargs
            )
            result = cache.get(cache_key)
            if result is None:
                result = func(*args, **kwargs)
                cache.set(cache_key, result, ttl)
            return result

        return wrapper

    return decorator_cache



import re
from datetime import timedelta

from django.contrib.staticfiles.storage import HashedFilesMixin, ManifestFilesMixin
from django.core.files.storage import get_storage_class
from google.oauth2 import service_account
from storages.backends.gcloud import GoogleCloudStorage
from storages.utils import setting


class StorageMixin:
    object_parameters = {
        "cache_control": f"max-age={60 * 60 * 24 * 7}, s-maxage={60 * 60 * 24 * 7}, must-revalidate"
    }
    default_acl = "publicRead"
    file_overwrite = False
    retry = True


class ManifestStaticFilesStorage(StorageMixin, ManifestFilesMixin, GoogleCloudStorage):
    bucket_name = setting("STATIC_BUCKET_NAME")
    custom_endpoint = setting("STATIC_CUSTOM_ENDPOINT")
    location = setting("STATIC_LOCATION", "")
    manifest_name = setting("STATIC_MANIFEST_NAME", ManifestFilesMixin.manifest_name)
    credentials = (
        service_account.Credentials.from_service_account_info(
            setting("STATIC_CREDENTIALS")
        )
        if setting("STATIC_CREDENTIALS")
        else None
    )
    vite_hash_pattern = r"^.+[\.-][0-9a-zA-Z_-]{8,12}\..+$"

    def url(self, name, force=False):
        # if the file already has a hash, we don't need the django hashed file
        if re.match(self.vite_hash_pattern, name):
            return super(HashedFilesMixin, self).url(name)
        return super().url(name, force)


class MediaStorage(StorageMixin, GoogleCloudStorage):
    bucket_name = setting("MEDIA_BUCKET_NAME")
    location = setting("MEDIA_LOCATION", "")
    custom_endpoint = setting("MEDIA_CUSTOM_ENDPOINT")
    credentials = (
        service_account.Credentials.from_service_account_info(
            setting("MEDIA_CREDENTIALS")
        )
        if setting("MEDIA_CREDENTIALS")
        else None
    )


class ThumbnailStorage(MediaStorage):
    file_overwrite = True


private_files_storage = get_storage_class()()

if hasattr(private_files_storage, "default_acl"):  # pragma: no cover
    private_files_storage.default_acl = "private"
    private_files_storage.expiration = timedelta(hours=1)
    private_files_storage.custom_endpoint = None



import os.path
from io import BytesIO

from django.core.files.base import ContentFile
from PIL import Image


def get_image(name="test.png", width=160, height=100):
    image = Image.new("RGB", (width, height), color="red")

    with BytesIO() as output:
        image.save(output, os.path.splitext(name)[1][1:])
        return ContentFile(output.getvalue(), name=name)



import os

from django.core.management.base import BaseCommand, OutputWrapper
from django.urls import get_resolver
from django.urls.resolvers import LocalePrefixPattern, RegexPattern
from django.utils.functional import Promise


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--output", default="/dev/stdout", help="Write to FILE instead of stdout"
        )

    def generate_urlpatterns(self, urls, indent=0):
        ws = "    " * indent
        for url in urls:
            if isinstance(url.pattern, LocalePrefixPattern):
                self.output.write(
                    """\
]

urlpatterns += i18n_patterns(
"""
                )
                self.generate_urlpatterns(url.url_patterns, 1)
                self.output.write(
                    """\
)

urlpatterns += ["""
                )
                continue

            if isinstance(url.pattern, RegexPattern):
                path = "re_path"
                pattern = (
                    f'_(r"{url.pattern}")'
                    if isinstance(url.pattern._regex, Promise)
                    else f'r"{url.pattern}"'
                )
            else:
                path = "path"
                pattern = (
                    f'_("{url.pattern}")'
                    if isinstance(url.pattern._route, Promise)
                    else f'"{url.pattern}"'
                )

            if hasattr(url, "url_patterns"):
                if url.namespace:
                    self.output.write(f"{ws}{path}({pattern}, include(([")
                    self.generate_urlpatterns(url.url_patterns, indent + 1)
                    self.output.write(f'{ws}], "{url.namespace}"))),')
                else:
                    self.output.write(f"{ws}{path}({pattern}, include([")
                    self.generate_urlpatterns(url.url_patterns, indent + 1)
                    self.output.write(f"{ws}])),")
            else:
                if url.name:
                    self.output.write(
                        f'{ws}{path}({pattern}, empty_view, name="{url.name}"),'
                    )
                else:
                    self.output.write(f"{ws}{path}({pattern}, empty_view),")

    def handle(self, *args, **options):
        if options["output"] == "/dev/stdout":
            self.output = self.stdout
            fd = None
        else:
            dir = os.path.dirname(options["output"])
            if dir and not os.path.exists(dir):
                os.makedirs(dir)
            fd = open(options["output"], "w")
            self.output = OutputWrapper(fd)

        self.output.write(
            """\
# fmt: off
from django.conf.urls.i18n import i18n_patterns  # noqa
from django.urls import include, path, re_path  # noqa
from django.utils.translation import gettext_lazy as _  # noqa


def empty_view(*args, **kwargs):
    pass


urlpatterns = ["""
        )
        self.generate_urlpatterns(get_resolver().url_patterns, 1)
        self.output.write("]")
        self.output.close()
        if fd:
            fd.close()



"""
    Copied and modified from django-js-routes https://github.com/ellmetha/django-js-routes

    JS Routes serializers
    =====================

    This modules defines serializer classes responsible for extracting Django URLs that should be
    exposed to the client side to a specific format (eg. JSON). The resoluting exports are likely to
    be used in a Javascript module to provide reverse lookup functionality on the client side.

"""

import json
import re
from urllib.parse import urljoin

from django.conf import settings
from django.urls import get_resolver
from django.urls.resolvers import (
    LocalePrefixPattern,
    RegexPattern,
    RoutePattern,
    URLPattern,
    URLResolver,
)

from .utils.text import replace


class URLPatternsSerializer:
    """The main class responsible for the serialization of the URLs exposed on the client side.

    This class implements a mechanism allowing to traverse the tree of URLs of a specific URL
    resolver in order to return the URLs that should be exposed on the client side in a specific
    format such as JSON. The exported object provides a mapping between fully qualified (namespaced)
    URL names and the corresponding paths.

    This class uses the default URL resolver if no one is set at initialization time.

    """

    _url_arg_re = re.compile(r"(\(.*?\))")
    _url_kwarg_re = re.compile(r"(\(\?P\<(.*?)\>.*?\))")
    _url_optional_char_re = re.compile(r"(?:\w|/)(?:\?|\*)")
    _url_optional_group_re = re.compile(r"\(\?\:.*\)(?:\?|\*)")
    _url_path_re = re.compile(r"<(.*?)>")

    def __init__(self, resolver=None):
        self.resolver = resolver or get_resolver()

    def to_json(self):
        """Serializes the URLs to be exported in a JSON array."""
        return json.dumps(dict(self._parse()))

    def _parse(self):
        return self._parse_resolver(self.resolver)

    def _parse_resolver(
        self, resolver, parent_namespace=None, parent_url_prefix=None, include_all=False
    ):
        namespace = ":".join(
            filter(lambda n: n, [parent_namespace, resolver.namespace])
        )
        include_all = include_all or (
            namespace in getattr(settings, "JS_ROUTES_INCLUSION_LIST", [])
        )

        urls = []
        for url_pattern in resolver.url_patterns:
            if isinstance(url_pattern, URLResolver):
                url_prefix = urljoin(
                    parent_url_prefix or "/", self._prepare_url_part(url_pattern)
                )
                urls = urls + self._parse_resolver(
                    url_pattern,
                    parent_namespace=namespace,
                    parent_url_prefix=url_prefix,
                    include_all=include_all,
                )
            elif isinstance(url_pattern, URLPattern) and url_pattern.name:
                url_name = ":".join(filter(lambda n: n, [namespace, url_pattern.name]))
                if (
                    url_name in getattr(settings, "JS_ROUTES_INCLUSION_LIST", [])
                    or include_all
                ):
                    full_url = self._prepare_url_part(url_pattern)
                    urls.append((url_name, urljoin(parent_url_prefix or "/", full_url)))

        return urls

    def _prepare_url_part(self, url_pattern):
        url = ""

        if isinstance(url_pattern.pattern, RegexPattern):
            url = url_pattern.pattern._regex
        elif isinstance(url_pattern.pattern, RoutePattern):
            url = url_pattern.pattern._route
        elif isinstance(url_pattern.pattern, LocalePrefixPattern):
            url = url_pattern.pattern.language_prefix
        else:  # pragma: no cover
            raise ValueError(
                'url_pattern must be a valid URL pattern ; "{}" is not'.format(
                    url_pattern
                )
            )

        final_url = replace(url, [("^", ""), ("$", "")])
        final_url = self._remove_optional_groups_from_url(final_url)
        final_url = self._remove_optional_characters_from_url(final_url)
        final_url = self._replace_arguments_in_url(final_url)

        return final_url

    def _remove_optional_groups_from_url(self, url):
        matches = self._url_optional_group_re.findall(url)
        return replace(url, [(el, "") for el in matches]) if matches else url

    def _remove_optional_characters_from_url(self, url):
        matches = self._url_optional_char_re.findall(url)
        return replace(url, [(el, "") for el in matches]) if matches else url

    def _replace_arguments_in_url(self, url):
        # Identifies and replaces named URL arguments inside the URL.
        kwarg_matches = self._url_kwarg_re.findall(url)
        url = (
            replace(url, [(el[0], "<{}>".format(el[1])) for el in kwarg_matches])
            if kwarg_matches
            else url
        )

        # Identifies and replaces unnamed URL arguments inside the URL.
        args_matches = self._url_arg_re.findall(url)
        url = replace(url, [(el, "<>") for el in args_matches]) if args_matches else url

        return url


url_patterns_serializer = URLPatternsSerializer()



"""
    Django JS Routes resolver dump command
    ======================================

    This command allows to export a JS module containing the Django URLs that should be exposed on
    the client side as of a resolver helper allowing to perform URL lookups.

"""

from os import path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.template.loader import render_to_string
from django.utils import translation

from moose_frank.js_routes.serializers import url_patterns_serializer


class Command(BaseCommand):
    """Dumps Django URLs and the resolver helper in a single TS file."""

    help = "Dump Django URLs and the resolver helper to a TS file per language."

    def add_arguments(self, parser):
        parser.add_argument(
            "-o",
            "--output-dir",
            required=True,
            help="Specifies a directory to which the output is written.",
        )

    def _write_resolver_file(self, filename):
        with open(filename, "w") as output:
            output.write(
                render_to_string(
                    "js_routes/_dump/resolver.ts",
                    {"routes": url_patterns_serializer.to_json()},
                )
            )

    def handle(self, *args, **options):
        filename = path.join(options["output_dir"], "index.ts")
        self._write_resolver_file(filename)

        for lang, _ in settings.LANGUAGES:
            with translation.override(lang):
                filename = path.join(options["output_dir"], f"{lang}.ts")
                self._write_resolver_file(filename)



import json
from typing import Optional


try:
    from strawberry_django_jwt.shortcuts import get_token
except ImportError:
    from graphql_jwt.shortcuts import get_token


class GraphQLMixin:
    def execute(
        self,
        query: str,
        variables: Optional[dict] = None,
        as_user: Optional = None,
        files: Optional[dict] = None,
        extra_token_data: Optional[dict] = None,
    ):
        headers = {}
        if as_user:
            if extra_token_data is None:
                extra_token_data = {}
            headers[
                "HTTP_AUTHORIZATION"
            ] = f"JWT {get_token(as_user, **extra_token_data)}"

        data = {
            "operations": json.dumps({"query": query, "variables": variables or {}})
        }
        if files:
            mapping = {i: [k] for i, k in enumerate(files.keys())}
            data["map"] = json.dumps(mapping)
            for i, keys in mapping.items():
                data[i] = files[keys[0]]
                if hasattr(data[i], "seek"):
                    data[i].seek(0)

        return self.client.post("/graphql", data=data, **headers)



"""
    JS Routes text utilities
    ========================

    This modules defines utilities for manipulating and performing operations on strings.

"""

from functools import reduce


def replace(data, replacements):
    """Allows to performs several string substitutions.

    This function performs several string substitutions on the initial ``data`` string using a list
    of 2-tuples (old, new) defining substitutions and returns the resulting string.

    """
    return reduce(lambda a, kv: a.replace(*kv), replacements, data)



from collections import OrderedDict

import graphene
from django.core.files.uploadedfile import UploadedFile
from django.forms import all_valid
from graphene import Field, InputField
from graphene.types.utils import yank_fields_from_attrs
from graphene_django.forms.mutation import DjangoModelDjangoFormMutationOptions
from graphene_django.forms.mutation import (
    DjangoModelFormMutation as BaseDjangoModelFormMutation,
)
from graphene_django.forms.mutation import fields_for_form
from graphene_django.forms.types import ErrorType
from graphene_django.registry import get_global_registry

from .types import ErrorTypeFormset, ErrorUnion


class DjangoModelFormMutation(BaseDjangoModelFormMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_object(cls, root, info, pk):
        return cls._meta.model._default_manager.get(pk=pk)

    @classmethod
    def get_form_kwargs(cls, root, info, **input):
        kwargs = {"data": input}
        pk = input.pop("id", None)
        if pk:
            kwargs["instance"] = cls.get_object(root, info, pk)
        return kwargs


class DjangoModelFormWithInlinesMutation(DjangoModelFormMutation):
    errors = graphene.List(ErrorUnion)

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        inlines=None,
        form_class=None,
        model=None,
        return_field_name=None,
        only_fields=(),
        exclude_fields=(),
        **options,
    ):
        if not inlines:
            raise Exception(
                "inlines are required for DjangoModelFormWithInlinesMutation"
            )

        if not form_class:
            raise Exception("form_class is required for DjangoModelFormMutation")

        if not model:
            model = form_class._meta.model

        if not model:
            raise Exception("model is required for DjangoModelFormMutation")

        form = form_class()
        input_fields = fields_for_form(form, only_fields, exclude_fields)
        input_fields["id"] = graphene.ID()

        for field, inline in dict(inlines).items():
            input_fields[f"{field}_total_forms"] = graphene.Int(required=True)
            input_fields[f"{field}_initial_forms"] = graphene.Int(required=True)

            inline_fields = fields_for_form(
                inline.form_class(), inline.fields or (), inline.exclude or ()
            )
            inline_fields["id"] = graphene.ID()
            inline_fields["delete"] = graphene.Boolean(default_value=False)

            inline_input = type(
                "{}Input".format(inline.__name__),
                (graphene.InputObjectType,),
                OrderedDict(yank_fields_from_attrs(inline_fields, _as=InputField)),
            )

            input_fields[field] = graphene.List(inline_input)

        registry = get_global_registry()
        model_type = registry.get_type_for_model(model)
        return_field_name = return_field_name
        if not return_field_name:
            model_name = model.__name__
            return_field_name = model_name[:1].lower() + model_name[1:]

        output_fields = OrderedDict()
        output_fields[return_field_name] = graphene.Field(model_type)

        _meta = DjangoModelDjangoFormMutationOptions(cls)
        _meta.inlines = inlines
        _meta.form_class = form_class
        _meta.model = model
        _meta.return_field_name = return_field_name
        _meta.fields = yank_fields_from_attrs(output_fields, _as=Field)

        input_fields = yank_fields_from_attrs(input_fields, _as=InputField)
        super(BaseDjangoModelFormMutation, cls).__init_subclass_with_meta__(
            _meta=_meta, input_fields=input_fields, **options
        )

    @classmethod
    def get_inlines(cls):
        return cls._meta.inlines

    @classmethod
    def construct_inlines(cls, request, obj, input):
        request.POST = request.POST.copy()

        inline_formsets = []
        for field, inline_class in dict(cls.get_inlines()).items():
            request.POST[f"{field}-TOTAL_FORMS"] = input[f"{field}_total_forms"]
            request.POST[f"{field}-INITIAL_FORMS"] = input[f"{field}_initial_forms"]

            for i, inline_data in enumerate(input.get(field, [])):
                for key, value in inline_data.items():
                    if key == "delete":
                        key = key.upper()

                    if isinstance(value, dict):
                        file = value.get("file", None)
                        if isinstance(file, UploadedFile):
                            request.FILES[f"{field}-{i}-{key}"] = file
                        if value.get("clear", False):
                            request.POST[f"{field}-{i}-{key}-clear"] = True
                    else:
                        request.POST[f"{field}-{i}-{key}"] = value

            inline_instance = inline_class(cls._meta.model, request, obj)
            inline_formset = inline_instance.construct_formset()
            inline_formsets.append(inline_formset)
        return inline_formsets

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        form = cls.get_form(root, info, **input)

        obj = form.instance
        if form.is_valid():
            obj = form.save(commit=False)
            form_validated = True
        else:
            form_validated = False

        inlines = cls.construct_inlines(info.context, obj, input)

        if all_valid(inlines) and form_validated:
            return cls.perform_mutate(form, inlines, info)
        else:
            errors = [
                ErrorType(field=key, messages=value)
                for key, value in form.errors.items()
            ]

            for inline in inlines:
                non_form_errors = inline.non_form_errors()
                if non_form_errors:
                    errors += [ErrorType(field=inline.prefix, messages=non_form_errors)]

                errors += [
                    ErrorTypeFormset(
                        field=inline.prefix,
                        form_index=i,
                        errors=[
                            ErrorType(field=key, messages=value)
                            for key, value in form.errors.items()
                        ],
                    )
                    for i, form in enumerate(inline.forms)
                    if not (inline.can_delete and inline._should_delete_form(form))
                    and form.errors
                ]

            return cls(errors=errors)

    @classmethod
    def perform_mutate(cls, form, inlines, info):
        obj = form.save()
        for formset in inlines:
            formset.save()
        kwargs = {cls._meta.return_field_name: obj}
        return cls(errors=[], **kwargs)



from typing import Iterable


def style_attribute_filter(allowed_styles: Iterable[str]):
    """
    A poor man's CSS filter for the nh3 sanitizing library.

    Example:

        The following example shows a style attribute being filtered. Note how `color` is
        filtered out, while `padding` is kept:

        >>> nh3.clean(
        ...     '<div style="color: red; padding: 10px">Hello</div>',
        ...     tags={"a", "div"},
        ...     attributes={
        ...         "a": {"href"},
        ...         "div": {"style"},
        ...     },
        ...     attribute_filter=style_attribute_filter({"border", "padding"}),
        ... )
        "<div style=\"padding:10px;\">Hello</div>"
    """

    def _filter(element, attribute, value):
        if attribute == "style":
            new_value = ""
            for line in value.split(";"):
                parts = line.split(":")
                if len(parts) != 2:
                    continue
                property, value = [x.strip() for x in parts]
                if property in allowed_styles:
                    new_value += f"{property}:{value};"
            return new_value
        return value

    return _filter



#!/usr/bin/env python

from setuptools import find_packages, setup


version = "9.3.0"

setup(
    name="moose-frank",
    packages=find_packages(),
    version=version,
    description="A Python package packed with tools that are commonly used in "
    "Moose projects.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Sven Groot (Moose / Digital Agency)",
    author_email="sven@wearemoose.com",
    url="https://gitlab.com/mediamoose/moose-frank/tree/v{}".format(version),
    download_url="https://gitlab.com/mediamoose/moose-frank/repository/v{}/archive.tar.gz".format(
        version
    ),
    include_package_data=True,
    install_requires=["django>=3.2.23"],
    extras_require={
        "graphene": [
            "graphene-django>=3.0",
            "graphene-file-upload>=1.3",
            "graphene>=3.2",
        ],
        "gcloud": [
            "django-storages[google]>=1.13.2",
        ],
    },
    license="MIT",
    zip_safe=False,
    keywords=["moose", "frank", "frankenstein"],
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



import operator
from typing import Optional

from django.core.exceptions import ValidationError
from django.core.files.images import get_image_dimensions
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class FileSizeValidator:
    message = _("Maximum file size is %(size).2g MB.")
    code = "invalid_file_size"

    def __init__(
        self,
        max_file_size: Optional[int] = None,
        message: Optional[str] = None,
        code: Optional[str] = None,
    ):
        self.max_file_size = max_file_size
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        if self.max_file_size is not None and value.size > self.max_file_size:
            raise ValidationError(
                self.message,
                code=self.code,
                params={"size": self.max_file_size / 1024 / 1024},
            )

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.max_file_size == other.max_file_size
            and self.message == other.message
            and self.code == other.code
        )


@deconstructible
class ImageDimensionsValidator:
    code = "invalid_image_dimensions"
    default_messages = {
        "lt": _("Image dimensions must be less than %(dimension)s."),
        "le": _("Image dimensions must be less than or equal to %(dimension)s."),
        "gt": _("Image dimensions must be greater than %(dimension)s."),
        "ge": _("Image dimensions must be greater than or equal to %(dimension)s."),
        "eq": _("Image dimensions must be equal to %(dimension)s."),
        "ne": _("Image dimensions may not be equal to %(dimension)s."),
    }

    def __init__(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        op: Optional[str] = "ge",
        message: Optional[str] = None,
        code: Optional[str] = None,
    ):
        self.width = width
        self.height = height
        self.operator = getattr(operator, op, operator.ge)
        self.message = message or self.default_messages[self.operator.__name__]
        if code is not None:
            self.code = code

    def __call__(self, value):
        if value.closed:
            value.open()
        width, height = get_image_dimensions(value)
        if (
            self.width
            and not self.operator(width, self.width)
            or self.height
            and not self.operator(height, self.height)
        ):
            raise ValidationError(
                self.message,
                code=self.code,
                params={"dimension": self._get_dimensions_format()},
            )

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.width == other.width
            and self.height == other.height
            and self.operator == other.operator
            and self.message == other.message
            and self.code == other.code
        )

    def _get_dimensions_format(self):
        if self.width and self.height:
            return _("%(width)dx%(height)d pixels") % {
                "width": self.width,
                "height": self.height,
            }
        elif self.width:
            return _("%(width)d pixels wide") % {"width": self.width}
        elif self.height:  # pragma: no cover
            return _("%(height)d pixels high") % {"height": self.height}
